{% extends "base.html" %}
{% load i18n %}

{% block title %}{% trans "Reset Password" %}{% endblock %}
{% block content %}
    <h1>{% trans "Reset Password" %}</h1>
    <form method="post">
        {% csrf_token %}
        {{ form.as_p }}
        <button type="submit">{% trans "Reset Password" %}</button>
    </form>
{% endblock %}
