{% extends "base.html" %}
{% load i18n %}

{% block title %}{% trans "Password Changed" %}{% endblock %}
{% block content %}
    <h1>{% trans "Password Changed" %}</h1>
    <p>{% trans "Your password has been changed successfully." %}</p>
    <p><a href="{% url 'accounts:login' %}">{% trans "Go to Login" %}</a></p>
{% endblock %}
